#include <openssl/bio.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>

#include "netex.h"
#include "config.h"

#define PORT_NUM_LENGHT 6

// https://invisible-island.net/ncurses/

void netex_init(void)
{
    // Curses
    initscr();
    noecho(); // Stop printing key presses
    if (has_colors() && can_change_color())
    {
        start_color();
        PRINT_LINE("%i colors available, %i color pairs available.", COLORS, COLOR_PAIRS);
        //TODO: Make color pairs
    }
    // OpenSSL
    SSL_load_error_strings();
    SSL_library_init();
    OpenSSL_add_all_algorithms();
}

void netex_shutdown(void)
{
    // OpenSSL
    ERR_free_strings();
    EVP_cleanup();
    // Curses
    endwin();
}

int setup_socket(const int port, const char* hostname, const SOCKET_IPV_TYPE ipv_type)
{
    HOST_INFO* resolved_host = resolve_host(hostname, port, ipv_type);
    if (resolved_host == NULL)
        return -1;
    int socket_fd;
    switch (resolved_host->ipv_type)
    {
        default:
        case IPv4:
            PRINT_LINE("Creating IPv4 socket, bound on: %s:%i", resolved_host->ip, port);
            socket_fd = socket(AF_INET, SOCK_STREAM, 0);
            if (socket_fd < 0)
            {
                ERROR("Failed to created socket (IPv4)");
                free_host_info(resolved_host);
                return -1;
            }
            break;
        case Double:
        case IPv6:
            PRINT_LINE("Creating IPv6 socket, bound on: [%s]:%i", resolved_host->ip, port);
            socket_fd = socket(AF_INET6, SOCK_STREAM, 0);
            if (socket_fd < 0)
            {
                ERROR("Failed to create socket (IPv6)");
                free_host_info(resolved_host);
                return -1;
            }
            break;
    }
    if (socket_fd == 0)
    {
        WARN("Failed to set socket!");
        free_host_info(resolved_host);
        return -1;
    }
    const int bind_result = bind(socket_fd, resolved_host->address, resolved_host->addrlen);
    if (bind_result != 0)
    {
        const int bind_error = errno;
        ERROR("Could not bind! Error: %i", bind_error);
        close(socket_fd);
    }
    free_host_info(resolved_host);
    return socket_fd;
}

HOST_INFO* resolve_host(const char* hostname, const int port, const SOCKET_IPV_TYPE ipv_type)
{
    struct addrinfo *resolved, *next, hints;

    const char* p_port = NULL;
    char port_str[PORT_NUM_LENGHT];
    if (port != 0)
    {
        const int print_result = snprintf(port_str, PORT_NUM_LENGHT, "%d", port);
        if (print_result < 0)
        {
            ERROR("Failed to convert port to string!");
            return NULL;
        }
        p_port = (char*)&port_str;
    }

    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_DGRAM;
    hints.ai_flags = AI_PASSIVE;
    hints.ai_protocol = 0;
    hints.ai_canonname = NULL;
    hints.ai_addr = NULL;
    hints.ai_next = NULL;

    switch (ipv_type)
    {
        default:
        case IPv4:
        hints.ai_family = AF_INET;
            break;
        case IPv6:
            hints.ai_family = AF_INET6;
            break;
        case Double:
            hints.ai_family = AF_UNSPEC;
            break;
    }
    if (getaddrinfo(hostname, p_port, &hints, &resolved) != 0)
    {
        ERROR("Failed to get addres info for hostname '%s'", hostname);
        return NULL;
    }
    for (next = resolved; next != NULL; next = next->ai_next)
    {
        const int sock_fd = socket(next->ai_family, next->ai_socktype, next->ai_protocol);
        if (sock_fd == -1)
            continue;
        if (bind(sock_fd, next->ai_addr, next->ai_addrlen) == 0)
            break;
        if (connect(sock_fd, next->ai_addr, next->ai_addrlen) == 0)
            break;
        close(sock_fd);
    }
    if (next == NULL)
    {
        freeaddrinfo(resolved);
        return NULL;
    }
    HOST_INFO* host_info = malloc(sizeof(HOST_INFO));
    if (host_info == NULL)
    {
        const int malloc_error = errno;
        ERROR("Failed to allocate object, error: %i", malloc_error);
        freeaddrinfo(resolved);
        return NULL;
    }
    switch (next->ai_family)
    {
        case AF_INET:
            host_info->ipv_type = IPv4;
            break;
        case AF_INET6:
            host_info->ipv_type = IPv6;
            break;
        default:
            ERROR("Invalid net family encounterd!");
            free(host_info);
            freeaddrinfo(resolved);
            return NULL;
    }
    char host[NI_MAXHOST];
    getnameinfo(next->ai_addr, next->ai_addrlen, host, sizeof(host), NULL, 0, NI_NUMERICHOST);
    host_info->ip = strdup(host);
    host_info->addrlen = next->ai_addrlen;
    host_info->address = malloc(next->ai_addrlen);
    if (host_info->address != NULL)
    {
        if (memcpy(host_info->address, next->ai_addr, next->ai_addrlen) == NULL)
        {
            WARN("Failed to copy addrinfo over!");
        }
    }
    freeaddrinfo(resolved);
    return host_info;
}

void free_host_info(HOST_INFO* host_info)
{
    if (host_info == NULL)
        return;
    if (host_info->address != NULL)
        free(host_info->address);
    if (host_info->ip != NULL)
        free(host_info->ip);
    free(host_info);
}